(**************************************************************************)
(*                                   Ott                                  *)
(*                                                                        *)
(*        Peter Sewell, Computer Laboratory, University of Cambridge      *)
(*      Francesco Zappa Nardelli, Moscova project, INRIA Rocquencourt     *)
(*                                                                        *)
(*  Copyright 2005-2008                                                   *)
(*                                                                        *)
(*  Redistribution and use in source and binary forms, with or without    *)
(*  modification, are permitted provided that the following conditions    *)
(*  are met:                                                              *)
(*  1. Redistributions of source code must retain the above copyright     *)
(*  notice, this list of conditions and the following disclaimer.         *)
(*  2. Redistributions in binary form must reproduce the above copyright  *)
(*  notice, this list of conditions and the following disclaimer in the   *)
(*  documentation and/or other materials provided with the distribution.  *)
(*  3. The names of the authors may not be used to endorse or promote     *)
(*  products derived from this software without specific prior written    *)
(*  permission.                                                           *)
(*                                                                        *)
(*  THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS    *)
(*  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     *)
(*  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE    *)
(*  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY       *)
(*  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL    *)
(*  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE     *)
(*  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS         *)
(*  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER  *)
(*  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR       *)
(*  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN   *)
(*  IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                         *)
(**************************************************************************)

exception NotImplementedYet;;

exception Rule_parse_error of string

open Types;;
open Location;;

(* library functions *) 

let pp_list_all m = 
  ( match m with 
  | Ascii _ | Tex _ -> Auxl.error "internal: pp_list_all for Ascii-Tex\n"
  | Twf _ -> Auxl.error "internal: pp_list_all for Twelf\n"
  | Caml _ -> Auxl.error "internal: pp_list_all for Caml\n"
  | Isa io -> Auxl.error "internal: pp_list_all for isa\n"
  | Hol ho -> Auxl.error "internal: pp_list_all for hol\n"
  | Coq co ->
      ( match co.coq_library with
      | None -> Auxl.error "internal: None found for coq_library - list_all\n"
      | Some r -> 
          let s,l = !r in
          if not (List.mem "list_all" l) then
          r :=
            ( s
              ^ "Definition list_all (A:Set) (p:A->Prop) :=\n"
              ^ "  fix list_all (l:list A) : Prop :=\n"
              ^ "  match l with\n"
              ^ "  | nil => True\n"
              ^ "  | cons a l' => p a /\\ list_all l'\n"
              ^ "end.\n"
              ^ "Implicit Arguments list_all.\n\n",
             ("list_all"::l))) )

let rn_formula="formula"

(** ******************************** *)
(** pretty printing of definitions   *)
(** ******************************** *)

let pp_subntr : pp_mode -> syntaxdefn -> (nontermroot * nontermroot * nonterm) -> string = 
  fun m xd -> 
    function (ntrl,ntru,(ntr',suff)) ->
      let s = 
	Auxl.pp_is ntrl ntru ^ " " 
	^ Auxl.hide_isa_trailing_underscore m
	    (( match m with Twf _ -> String.uppercase ntr' 
	    | Coq _ | Isa _ | Hol _ -> ntr'
	    | Caml _ | Tex _ | Ascii _ -> raise Auxl.ThisCannotHappen )
	     ^ Grammar_pp.pp_suffix_with_sie m xd Bounds.sie_project suff)
      in ( match m with
      | Coq co when not co.coq_expand_lists -> "Is_true ("^s^")"
      | _ -> s )
          
let pp_listsubntr : pp_mode -> syntaxdefn -> ((nontermroot * nontermroot * nonterm) list * string * string * string) list -> string list =
  fun m xd listsubntrs -> 
    let clauses = 
      List.flatten
        (List.map 
           (function (subntrs,pp_squished_vars,pp_pattern,coq_type_pattern) ->
             List.map
               (function subntr ->
                 match m with
                 | Ascii _ | Tex _ -> Auxl.error "internal: pp_listsubntr of Tex-Ascii\n"
                 | Caml _ ->
                     Auxl.error "internal: defns not implemented for OCaml"
                 | Twf _ ->
                     Auxl.error "internal: defns not implemented for Twelf"
                 | Isa _ -> 
(* Tobias says: should use  "ALL x : set xs. P x" rather than "list_all P xs", so...: *)
(* TODO: choose a better internal name than x__ *)
                     "List.list_all "
                     ^ "(%"^pp_pattern^". "^pp_subntr m xd subntr^") "
                     ^ pp_squished_vars
                         
(*                            "(ALL x__ : set "^pp_squished_vars *)
(*                            ^ ".  let "^pp_pattern^"=x__ in " *)
(*                            ^ pp_subntr m xd subntr^")" *)
                              
                 | Hol _ -> 
                     "EVERY "
                     ^ "(\\"^pp_pattern^". "^pp_subntr m xd subntr^") "
                     ^ pp_squished_vars

                 | Coq co -> 
	             let ty_list = Str.split (Str.regexp "(\\|*\\|)") coq_type_pattern in
                     let list_all_name =
                       if co.coq_expand_lists
                       then (pp_list_all m; "list_all")
                       else "Forall_list"
                     in
	             if String.contains coq_type_pattern '*' 
 	             then begin
                       "(" ^ list_all_name ^ " (fun (tmp_:"^coq_type_pattern^") => match tmp_ with "
		       ^ pp_pattern ^ " => "
		       ^ pp_subntr m xd subntr ^ " end) "
		       ^ ( if co.coq_expand_lists 
		           then "(unmake_list_"^(String.concat "_" ty_list)^" "^ pp_squished_vars^")"
		           else pp_squished_vars )
		       ^ ")"
	             end else begin
	               "(" ^ list_all_name ^ " (fun ("^pp_pattern^":"^coq_type_pattern^") => "  (* FZ too many parentheses *)
		       ^ pp_subntr m xd subntr ^ ") "
		       ^ ( if co.coq_expand_lists 
		           then "(unmake_list_"^(String.concat "_" ty_list)^" "^ pp_squished_vars^")"
		           else pp_squished_vars )
		       ^ ")"
	             end)
               subntrs)
           listsubntrs) in
    clauses

(* this function collapses (e1:term) (e2:term) into (e1 e2:term) *)
(* it is used when pp Coq forall quantifications in definitions *)
let rec collapse_quantifications l =
  let rec collapse b t l =
  match b, t, l  with 
  | None, None, [] -> []
  | Some b1, Some t1, [] -> [(b1,Some t1)]
  | None, None, (v,Some t)::tl -> collapse (Some [v]) (Some t) tl
   | None, None, (v,None)::tl -> ([v],None)::(collapse None None tl) 
  | Some b1, Some t1, (v, Some t)::tl -> 
      if String.compare t t1 = 0 
      then collapse (Some (b1@[v])) (Some t1) tl
      else (b1,Some t1)::(collapse (Some [v]) (Some t) tl) 
  | Some b1, Some t1, (v, None)::tl -> 
      (b1,Some t1)::([v],None)::(collapse None None tl)  in 
  collapse None None l
                 
          
let pp_drule (m:pp_mode) (xd:syntaxdefn) (dr:drule) : string =
(*   print_endline ("&*"^dr.drule_name); *)
(*   print_endline (String.concat "\n" (List.map Grammar_pp.pp_plain_symterm dr.drule_premises));  *)
(*   print_endline (Grammar_pp.pp_plain_symterm dr.drule_conclusion);  *)

  let (((de1:dotenv1),de2) as de,de3,pptbe) = 
    Bounds.bound_extraction m xd dr.drule_loc (dr.drule_conclusion::dr.drule_premises) in

  (* FZ hack to get the right quantification when pp formula_dots in Coq using native lists *)
  ( match m with
  | Coq co when not co.coq_expand_lists ->
      co.coq_quantified_vars_from_de := 
	List.map (fun (x,_,_) -> x)
	  (Grammar_pp.extract_quantified_proof_assistant_vars m xd de1 de2 de3);
  | _ -> () );
  
  (* print_endline ("rule: " ^dr.drule_name); *)

  let ppd_premises = 
    List.map 
      (fun p -> (* print_endline (Grammar_pp.pp_plain_symterm p); *) Grammar_pp.pp_symterm m xd [] de p) 
      dr.drule_premises in
  let ppd_conclusion = (* print_endline (Grammar_pp.pp_plain_symterm dr.drule_conclusion); *) Grammar_pp.pp_symterm m xd [] de dr.drule_conclusion in
  match m with
  | Ascii ao ->
      (* first find out the width, without colorization *)
      let m0 = Ascii {ao with ppa_colour=false } in
      let ppd_premises0 = List.map 
          (Grammar_pp.pp_symterm m0 xd [] de) 
          dr.drule_premises in
      let ppd_conclusion0
          = Grammar_pp.pp_symterm m0 xd [] de dr.drule_conclusion in
      let length = List.fold_left max 0 
          (List.map String.length (ppd_conclusion0::ppd_premises0)) in
      (* now really pp *)
      let line = 
        (String.make length '-' ) ^ " " 
        ^ Grammar_pp.pp_COLONCOLON ^ dr.drule_name in
      pptbe ^
      (String.concat "\n" ppd_premises) ^ "\n" 
      ^ line ^ "\n"
      ^ ppd_conclusion ^ "\n"
  | Tex _ ->                                  
      let pp_com = 
        (match Auxl.hom_spec_for_hom_name "com" dr.drule_homs
        with
        | None -> ""
        | Some hs ->
            "\\text{"
            ^ String.concat "" (Grammar_pp.apply_hom_spec m xd hs [])
            ^ "}") in
      "\\newcommand{"^Grammar_pp.tex_drule_name m dr.drule_name^"}[1]{"
      ^ Grammar_pp.pp_tex_DRULE_NAME m ^ "[#1]{%\n"
      ^ String.concat "" (List.map (fun p-> ""^Grammar_pp.pp_tex_PREMISE_NAME m^"{"^p^"}%\n") ppd_premises)
      ^ "}{\n"
      ^ ppd_conclusion
      ^ "}{\n"
      ^ "{" ^ Grammar_pp.pp_tex_DRULE_NAME_NAME m ^"{"^Auxl.pp_tex_escape dr.drule_name^"}}"
      ^ "{"^pp_com^"}"
      ^ "\n}}\n"

(*        ^ "\\frac" *)
(*         ^" { \\begin{array}{l}" *)
(*         ^ ( String.concat "\\\\ \n" ppd_premises) *)
(*         ^ "\\end{array} }\n { " *)
(*         ^ ppd_conclusion *)
(*         ^ " } \\quad " *)
(*         ^ "\\textsc{"^Auxl.pp_tex_escape dr.drule_name^"}"^"\\quad" *)
(*         ^ pp_com *)
(*         ^ "\n \\]}\n" *)
  | Isa _ | Hol _ | Coq _ | Twf _ ->
      let non_free_ntrs = Subrules_pp.non_free_ntrs m xd in

      (* find all the nonterms used at non-free types *)
      let subntrs_of_ntmvsns : (nt_or_mv*subntr_data) list -> (nontermroot * nontermroot * nonterm) list = 
        function ntmvsns ->
          Auxl.option_map 
            (function x -> match x with
            | (Ntr ntr,suffix),(None as subntr_data) -> 
                (try  
                  let ntr_upper = List.assoc ntr non_free_ntrs in
                  Some (ntr,ntr_upper,(ntr,suffix))
                with
                  Not_found -> None)
            | (Ntr ntr,suffix),(Some(ntrl,ntru) as subntr_data) -> Some (ntrl,ntru,(ntr,suffix))
            | (Mvr mvr,suffix),_ -> None) 
            ntmvsns in

      (* first the non-list nonterms used at non-free types*)
      let nonlist_subntrs = subntrs_of_ntmvsns de3 in

      (* and second the list nonterms used at non-free types *)
      let list_subntrs :((nontermroot * nontermroot * nonterm) list * string * string * string) list = 
        List.map  
          (function (bound,de1i (*(ntmvsns,pp_squished_vars,pp_pattern,coq_type_pattern,hol_type_var)*)) ->
            subntrs_of_ntmvsns de1i.de1_ntmvsns,de1i.de1_compound_id,de1i.de1_pattern,de1i.de1_coq_type_of_pattern)
          de1 in

      let ppd_subntrs = 
        List.map (pp_subntr m xd) nonlist_subntrs 
        @ pp_listsubntr m xd list_subntrs in

      (* collect all the isa/coq/hol variables that should be quantified *)
      (* for this clause *)

      let quantified_proof_assistant_vars = 
	Grammar_pp.extract_quantified_proof_assistant_vars m xd de1 de2 de3 in

      (* calculate list length constraints *)
      (*   There are several different things we could do here: *)
      (*   -a- For a proof assistant list arising from ott list forms with bounds that      *)
      (*        include an upper bound and (perhaps implicitly) a lower bound               *)
      (*        (i.e. Bound_dotform, Bound_comp_u, or Bound_comp_lu, but not Bound_comp),   *)
      (*        say 1..n, we could add a premise that the length of the proof assistant list*)
      (*        is n.                                                                       *)
      (*    -a.2-  It might be a nice optimisation to only include that premise if the n    *)
      (*        is also mentioned elsewhere in the drule.                                   *)
      (*   -b- For a proof assistant list arising from ott list forms with bounds that      *)
      (*        have a number of dots (i.e. Bound_dotform or Bound_comp_lu), say 1....n,    *)
      (*        we could add a premise that the length of the proof assistant list is >= 2. *)
      (*   -c- We could record in each Ste_list any dot length constraint from the ott      *)
      (*        source grammar, then add up the length constraints from the                 *)
      (*        symterm_list_items comprising the Ste_list, see if a non-trivial additional *)
      (*        constraint is required to make the Ste_list of length at least that in the  *)
      (*        grammar, and, if so, add that as a premise.                                 *)
      (*   -d- For each projection index, eg ti, we could add a constraint that i is < the  *)
      (*        length of the proof assistant list.                                         *)
      (*  *)
      (* Below is code to do -a-, fine for isa & hol, but not enough for Coq as there       *)
      (*  we need to add quantifiers for the vars that occur only in bounds                 *)
(*         let list_vars_with_bounds = List.map (function (b,(_,compound_ident,_,_,_)) -> compound_ident,b) de1 in *)
(*         let si_subtract si_lower si_upper = match si_lower,si_upper with *)
(*         | Si_num s, Si_var (suffv,n) ->  *)
(*             let n' = n + 1 - int_of_string s in *)
(*             if n'=0 then suffv else (match m with *)
(*             | Isa _ -> "("^suffv^"+("^string_of_int n'^"::nat))" *)
(*             | Hol _ -> "("^suffv^"+"^string_of_int n'^")" *)
(*             | Coq _ -> "("^suffv^"+"^string_of_int n'^")") *)
(*         | _,_ -> failwith ("internal error: surprising suffixes in si_subtract: "^Grammar_pp.pp_plain_suffix_item si_lower^" and "^Grammar_pp.pp_plain_suffix_item si_lower) in *)
(*         let lengths b = match b with *)
(*         | Bound_dotform bd -> Some (si_subtract bd.bd_lower bd.bd_upper), Some bd.bd_length *)
(*         | Bound_comp bc -> None,None *)
(*         | Bound_comp_u bcu -> Some (si_subtract (Si_num "0") bcu.bcu_upper), None *)
(*         | Bound_comp_lu bclu -> Some (si_subtract bclu.bclu_lower bclu.bclu_upper), Some bclu.bclu_length in *)
(*         let list_length_constraints = Auxl.option_map (function  (b,(_,compound_ident,_,coq_type_pattern,_)) ->  *)
(*           match lengths b with *)
(*           | Some s, _ ->  *)
(*               Some ((match m with *)
(*               | Isa _ -> "length" *)
(*               | Hol _ -> "LENGTH" *)
(*               | Coq _ ->  *)
(*                   let ty_list = Str.split (Str.regexp "(\\|*\\|)") coq_type_pattern in *)
(*                   "length_list_"^(String.concat "_" ty_list) *)
(*                   (\* "[[TODO: insert appropriate Coq list length function for the compound_ident, built from the coq_type_pattern, here ]]" *\)) *)
(*               ^ " " ^ compound_ident^" = "^s) *)
(*           | None,_ -> None) de1 in *)

(*         let ppd_premises = ppd_premises @ list_length_constraints in *)
      
      ( match m with
      | Isa _ ->
          dr.drule_name^"I"
          ^ ": "  (* todo - should we add the I here or before? *)
          ^ "\"" 
          ^ 
            (if ppd_premises=[] && ppd_subntrs=[]  then "" 
            else 
	      "[|"
	      ^ (String.concat " ;\n" (ppd_subntrs @ ppd_premises)) 
	      ^ "|] ==> \n") 
          ^ ppd_conclusion 
          ^ "\"" ^ "\n"
      | Hol _ ->
          let clause_name_pp = "(clause_name \"" ^ dr.drule_name^ "\")" in
          "(" 
          ^ " (* "^dr.drule_name^" *)  "
          ^ 
            (match quantified_proof_assistant_vars with
            | [] -> ""
            | _ -> "!" 
                ^ (let quantified_proof_assistant_vars' = 
		  List.map 
		    (function (var,ty,_) -> "("^var^":"^ty^")") 
		    quantified_proof_assistant_vars in
                String.concat " " quantified_proof_assistant_vars' ^ " . "))
          ^ 
            (if ppd_premises=[] && ppd_subntrs=[] then
              clause_name_pp^" ==> \n" 
            else 
              clause_name_pp^" /\\\n"
              ^ "("
	      ^ (String.concat " /\\\n" 
                   (List.map (function s->"("^s^")") (ppd_subntrs @ ppd_premises)) )
	      ^ ")\n ==> \n") 
          ^ "("^ppd_conclusion ^")"
          ^ ")\n\n"

      | Coq co when not co.coq_expand_lists -> 

	  let nlh, nlhv = !(co.coq_non_local_hyp_defn), !(co.coq_non_local_hyp_defn_vars) in
	  co.coq_non_local_hyp_defn := "";
	  co.coq_non_local_hyp_defn_vars := [];

	  let qpavt_list =
	    let t = (Auxl.remove_duplicates ((List.map (fun (x,_,(t,_)) -> (x,Some t)) quantified_proof_assistant_vars) @ nlhv)) in  
	    (* Arthur insists that for locally nameless it is important to have the (L:vars) quantifications at the beginning *)
	    let (t1,t2) = List.partition (fun (v,t) -> match t with Some t1 -> String.compare t1 "vars" = 0 | None -> false) t in
	    collapse_quantifications (t1 @ t2) 
	  in

          let qpavt = 
	    List.map 
	      (fun (vs,t) -> 
		match t with
		| Some t -> "("^(String.concat " " vs)^":"^t^")"
		| None -> List.hd vs) (* if there are no type annotation, vs is a singleton *)
	      qpavt_list in

 	  "\n | " ^ dr.drule_name ^ " : " 
	  ^ ( if qpavt = [] then "\n"
              else ( "forall "
                     ^ String.concat " " qpavt ^ ",\n") )
	  ^ ( if ppd_premises = [] && ppd_subntrs = [] && (String.compare nlh "" = 0) 
          then ""
          else begin
            nlh ^ 
	    ( if (List.length (ppd_subntrs @ ppd_premises) = 0)
	    then ""
	    else "     " ^ (String.concat " ->\n     " (ppd_subntrs @ ppd_premises) ^ " ->\n") ) 
	  end )
	  ^ "     " ^ ppd_conclusion 

      | Coq co when co.coq_expand_lists -> 

          let required_list_types = 
            List.map (function (_,de1i) -> 
              List.map (function ((ntmvr,_),_) -> 
                Auxl.promote_ntmvr xd
                  (Auxl.primary_nt_or_mv_of_nt_or_mv xd ntmvr)) de1i.de1_ntmvsns) de1 in
          
          let missing_list_types = 
            List.filter 
              (fun ss -> 
                let name = String.concat "_" (List.map (fun s -> Grammar_pp.pp_nt_or_mv_root m xd (Auxl.promote_ntmvr xd s)) ss) in
                not (List.mem ("list_"^name) !(co.coq_list_types))) 
              required_list_types in

(*            print_endline ("rule: " ^dr.drule_name); *)
(*             print_endline ("missing: "^ (String.concat " -- " *)
(*                                            (List.map *)
(*                                               (Grammar_pp.pp_nt_or_mv_root m xd) *)
(*                                               (List.flatten missing_list_types)))); *)

          let (new_list_names, new_list_rules) = 
            List.split (List.map Transform.create_list_rule missing_list_types) in
          
          let rec add_list_rules_defns nln nlr =
            ( match (nln,nlr) with 
            | [],[] -> ()
            | h1::t1,h2::t2 -> 
                if not (List.mem h1 !(co.coq_list_types)) 
                then begin
                  co.coq_list_types := h1::!(co.coq_list_types);
                  co.coq_list_rules_defns := h2::!(co.coq_list_rules_defns);
                end;
                add_list_rules_defns t1 t2
	    | _,_ -> raise Auxl.ThisCannotHappen ) in
          
          add_list_rules_defns new_list_names new_list_rules;

	  let nlh, nlhv = !(co.coq_non_local_hyp_defn), !(co.coq_non_local_hyp_defn_vars) in
	  co.coq_non_local_hyp_defn := "";
	  co.coq_non_local_hyp_defn_vars := [];
	  let vtl = (List.map (fun (x,_,(t,_)) -> (x,Some t)) quantified_proof_assistant_vars) @ nlhv in
	  let vtlr = 
	    let rec check_el el l =
	      List.mem el (List.map fst l) in
	    let rec remove_dupl l =
	      ( match l with
	      | [] -> []
	      | (v,t)::tl -> 
		  if check_el v tl 
		  then remove_dupl tl
		  else (v,t)::(remove_dupl tl) ) in
	    remove_dupl vtl in
	  
	  let qpavt_list =
	    (* Arthur insists that for locally nameless it is important to have the (L:vars) quantifications at the beginning *)
	    let (t1,t2) = List.partition (fun (v,t) -> match t with Some t1 -> String.compare t1 "vars" = 0 | None -> false) vtlr in
	    collapse_quantifications (t1@t2)
	  in
	  
	  
          let qpavt = 
	    List.map 
	      (fun (vs,t) -> 
		match t with
		| Some t -> "("^(String.concat " " vs)^":"^t^")"
		| None -> List.hd vs)
	     qpavt_list in
(*	    (Auxl.remove_duplicates ((List.map (fun (x,_,(t,_)) -> (x,Some t)) quantified_proof_assistant_vars) @ nlhv)) in *)
 	  "\n | " ^ dr.drule_name ^ " : " ^ 
          ( if qpavt = [] then "\n"
          else ( "forall "
                 ^ String.concat " " qpavt ^ ",\n") )

	  ^ ( if ppd_premises = [] && ppd_subntrs = [] && (String.compare nlh "" = 0) 
          then ""
          else begin
            nlh ^ 
	    ( if (List.length (ppd_subntrs @ ppd_premises) = 0)
	    then ""
	    else "     " ^ (String.concat " ->\n     " (ppd_subntrs @ ppd_premises) ^ " ->\n") ) 
	  end )
	  ^ "     " ^ ppd_conclusion 

      | Twf wo ->
          !(wo.twf_current_defn) ^"/"^ dr.drule_name
          ^ ": "
	  ^ ppd_conclusion 
          ^ "" 
          ^ (if ppd_premises=[] && ppd_subntrs=[] then "" 
             else " <- " ^ (String.concat " <- " (ppd_subntrs @ ppd_premises)))
          ^ "."

      | Tex _ | Ascii _ -> raise Auxl.ThisCannotHappen 
      | Caml _ -> raise Auxl.ThisCannotHappen
       )
  | Caml oo ->
      Auxl.error "internal: defns not implemented for OCaml"
  | Twf _ ->
      Auxl.error "internal: defns not implemented for Twelf"


let pp_defncom (m:pp_mode) (xd:syntaxdefn) lookup (es:embed_spec) : string =
  match m with
  | Tex _  ->
      (Embed_pp.pp_embed_spec m xd lookup es)
  | _ -> ""

let pp_processed_semiraw_rule (m:pp_mode) (xd:syntaxdefn) (psr:processed_semiraw_rule) : string option =
 match psr with
    | PSR_Rule dr -> Some (pp_drule m xd dr)
    | PSR_Defncom es -> None

let pp_defn (m:pp_mode) (xd:syntaxdefn) lookup (defnclass_wrapper:string) (d:defn) : string =
  match m with
  | Ascii _ ->
      Grammar_pp.pp_DEFN ^"\n"
      ^ Grammar_pp.pp_symterm m xd [] de_empty d.d_form
      ^ Grammar_pp.pp_COLONCOLON ^" "^d.d_wrapper^" "
      ^ Grammar_pp.pp_BY
      ^ "\n\n"
      ^ String.concat "\n" (Auxl.option_map (pp_processed_semiraw_rule m xd) d.d_rules)
(*   | Isa io when not io.ppi_isa07_syntax -> *)
(*       "(\* defn "^d.d_name^" *\)\n\n"  *)
(*       ^ String.concat "\n" (Auxl.option_map (pp_processed_semiraw_rule m xd) d.d_rules) *)
  | Isa io ->
      "(* defn "^d.d_name^" *)\n\n" 
      ^ ( match (Grammar_pp.isa_fancy_syntax_clause_for_defn m xd io "foo" defnclass_wrapper d) with
        | None -> ""
	| Some _ -> "| " ) 
      ^ String.concat "\n| " (Auxl.option_map (pp_processed_semiraw_rule m xd) d.d_rules)
  | Hol _ ->
      "(* defn "^d.d_name^" *)\n\n" 
      ^ String.concat "/\\ " (Auxl.option_map (pp_processed_semiraw_rule m xd) d.d_rules)
  | Coq co -> (* FZ factor this code ? *)
 (*      print_endline ("************"^Grammar_pp.pp_plain_symterm d.d_form);  (\* FZ REMOve *\) *)
(*       print_endline (Grammar_pp.pp_syntaxdefn pp_ascii_opts_default xd);  (\* FZ DEBUG *\) *)

      let prod_name = defnclass_wrapper ^ d.d_name in

      let type_defn = 
        let es = (Auxl.prod_of_prodname xd prod_name).prod_es in
        let ss = (Auxl.option_map (Grammar_pp.pp_element m xd []) es) in
        match ss with
        | [] -> "Prop :="
        | [s] -> s ^ " -> Prop :="      
        | _ -> String.concat " -> " ss ^" -> Prop :=" in        
      let defn = 
	defnclass_wrapper ^ d.d_name ^ " : "^type_defn ^"    "^ "(* defn "^d.d_name^" *)"
        ^ String.concat "" (Auxl.option_map (pp_processed_semiraw_rule m xd) d.d_rules)  in
      defn
  | Twf wo ->
      wo.twf_current_defn := d.d_name;
      "%% defn "^d.d_name^"\n\n" 
      ^ String.concat "\n" (Auxl.option_map (pp_processed_semiraw_rule m xd) d.d_rules)

  | Tex xo ->
      let prod_name = (defnclass_wrapper^d.d_name) in
      let p = Auxl.prod_of_prodname xd prod_name in
      let es = p.prod_es in 
      let pp_com = Grammar_pp.pp_com_es m xd d.d_homs es  in
      "% defn "^d.d_name^"\n"
      ^ String.concat "\n\n" (Auxl.option_map (pp_processed_semiraw_rule m xd) d.d_rules)
      ^ "\n\\newcommand{"^Grammar_pp.tex_defn_name m defnclass_wrapper d.d_name^"}[1]{"
      ^ "\\begin{"^Grammar_pp.pp_tex_DEFN_BLOCK_NAME m ^"}[#1]{$" 
      ^ Grammar_pp.pp_symterm m xd [] de_empty d.d_form 
      ^ "$}{"
      ^ pp_com
      ^ "}\n"
      ^ String.concat "" 
          (List.map 
             (function psr ->
               (match psr with
               | PSR_Rule dr -> 
                   let categories =
                     if (xo.ppt_show_categories &&
                         not (StringSet.is_empty dr.drule_categories)) then
                       "\\quad\\textsf{[" ^
                       String.concat " " (StringSet.fold (fun x l -> x::l)
                                            dr.drule_categories []) ^
                       "]}"
                     else ""
                   in
                   String.concat "" [Grammar_pp.pp_tex_USE_DRULE_NAME m;
                                     "{";
                                     Grammar_pp.tex_drule_name m dr.drule_name;
                                     "{}";
                                     categories;
                                     "}\n"]
               | PSR_Defncom es -> pp_defncom m xd lookup es))
             d.d_rules)
      ^ "\\end{"^Grammar_pp.pp_tex_DEFN_BLOCK_NAME m ^"}" 
      ^ "}\n\n"
  | Caml oo ->
      Auxl.error "internal: defns not implemented for OCaml"

            
let pp_defnclass (m:pp_mode) (xd:syntaxdefn) lookup (dc:defnclass) : string =
  let isa_type_of_defn : syntaxdefn -> defn -> string = 
    fun xd d ->  
      (* seems simplest to find the type associated with the production that
         we added to the language for this defn, rather than build a type
         directly from the defn. Though (now wrappers are working properly)
         that does mean we need to pass them in. *)
      let prod_name = dc.dc_wrapper ^ d.d_name in
      let es = (Auxl.prod_of_prodname xd prod_name).prod_es in
      let ss = (Auxl.option_map (Grammar_pp.pp_element m xd []) es) in
      match ss with
      | [] -> "unit"
      | [s] -> s      
      | _ -> "("^String.concat "*" ss ^")" in

  match m with
  | Ascii _ -> 
      Grammar_pp.pp_DEFNCLASS ^"\n"
      ^ dc.dc_name ^" "
      ^ Grammar_pp.pp_COLONCOLON^" "^dc.dc_wrapper^" "
      ^ Grammar_pp.pp_CCE^"\n\n"
      ^ String.concat "\n\n" (List.map (pp_defn m xd lookup dc.dc_wrapper) dc.dc_defns)

(*   old isa05 syntax *)
(*   | Isa io when not io.ppi_isa07_syntax -> *)
(*       "(\*" ^ "defns "^dc.dc_name^" *\)\n" *)
(*       ^ "consts\n  " *)
(*       ^ (String.concat  *)
(*            "\n  "  *)
(*            (List.map  *)
(*               (function d ->  *)
(*                 dc.dc_wrapper^d.d_name *)
(*                 ^ " :: \"" *)
(*                   ^ isa_type_of_defn xd d *)
(*                 ^ " set" ^"\"")  *)
(*               dc.dc_defns))  *)
(*       ^ "\n\n" *)
(*       ^ (let fancy_syntax_clauses =  *)
(*           Auxl.option_map  *)
(* 	    (Grammar_pp.isa_fancy_syntax_clause_for_defn m xd io dc.dc_name dc.dc_wrapper)  *)
(* 	    dc.dc_defns in *)
(*         match fancy_syntax_clauses with *)
(*         | [] -> "" *)
(*         | fscs ->  *)
(*             "syntax\n" *)
(*             ^ String.concat "\n" (List.map fst fscs) *)
(*             ^ "\n\ntranslations\n" *)
(*             ^ String.concat "\n" (List.map snd fscs) *)
(*             ^ "\n\n") *)
(*       ^ "inductive " *)
(*       ^ (String.concat  *)
(*            " "  *)
(*            (List.map (function d -> dc.dc_wrapper ^ d.d_name) dc.dc_defns))  *)
(*       ^ "\nintros\n\n" *)
      ^ (String.concat "\n" (List.map (pp_defn m xd lookup dc.dc_wrapper) dc.dc_defns))

  | Isa io ->
      let fancy_syntax_clauses = 
        Auxl.option_map
	  (Grammar_pp.isa_fancy_syntax_clause_for_defn m xd io dc.dc_name dc.dc_wrapper)
	  dc.dc_defns in

      "(*" ^ "defns "^dc.dc_name^" *)\n"
      ^ "inductive_set "
      ^ (String.concat 
           "\n and " 
           (List.map 
              (function d -> 
                dc.dc_wrapper^d.d_name
                ^ " :: \""
                  ^ isa_type_of_defn xd d
                ^ " set" ^"\"") 
              dc.dc_defns)) 
      ^ "\n"
      ^ ( match fancy_syntax_clauses with
          | [] -> ""
          | fscs -> "and" ^ String.concat "\n" (List.map fst fscs) ^"\n" )
      ^ "where\n"
      ^ ( match fancy_syntax_clauses with
          | [] -> ""
          | fscs -> String.concat "\n" (List.map snd fscs) ^"\n" )
      ^ (String.concat "\n| " (List.map (pp_defn m xd lookup dc.dc_wrapper) dc.dc_defns))

  | Hol ho -> 
      "(* defns "^dc.dc_name^" *)\n\n"
      ^ "val ("^ dc.dc_name^"_rules" 
      ^ ", "   ^ dc.dc_name^"_ind"
      ^ ", "   ^ dc.dc_name^"_cases"
      ^ ") = Hol_reln `\n"
      ^ (String.concat "/\\\n" 
	   (List.map (pp_defn m xd lookup dc.dc_wrapper) dc.dc_defns))
      ^ "`;\n"

  | Coq co -> 
      let defns_list = 
        String.concat "\nwith " 
	  (List.map (pp_defn m xd lookup dc.dc_wrapper) dc.dc_defns) in
      let list_aux_defns = !(co.coq_list_aux_defns.newly_defined) in
      co.coq_list_aux_defns.newly_defined := [];
      "\n(* defns "^dc.dc_name^" *)\n"
      ^ "Inductive "
      ^ defns_list
      ^ ( if list_aux_defns = [] 
          then "." 
          else "\n" ^ (String.concat "\n" list_aux_defns) ^ "." )

  | Twf wo -> 
      let twf_type_of_defn : syntaxdefn -> defn -> string = 
        fun xd d ->  
        (* follows isa case here *)
          let prod_name = dc.dc_wrapper ^ d.d_name in
          let es = (Auxl.prod_of_prodname xd prod_name).prod_es in
          let ss = (Auxl.option_map (Grammar_pp.pp_element m xd []) es) in
          match ss with
          | [] -> "unit"  (* FZ CHECK THIS *)
          | [s] -> s      
          | _ -> String.concat " -> " ss in
      
      let defns_list = 
        String.concat "\n" 
	  (List.map (pp_defn m xd lookup dc.dc_wrapper) dc.dc_defns) in
(*      let list_aux_defns = !(wo.twf_list_aux_defns.newly_defined) in *)
(*      wo.twf_list_aux_defns.newly_defined := []; *)
      "%% defns "^dc.dc_name^"\n\n"
      ^ (String.concat 
           "\n  " 
           (List.map 
              (function d -> 
                dc.dc_wrapper^d.d_name
                ^ " : "
                ^ twf_type_of_defn xd d  ^ " -> type.") 
              dc.dc_defns)) 
      ^ "\n\n"
      ^ defns_list
(*      ^ (String.concat "\n" list_aux_defns) *)

  | Caml oo ->
      Auxl.error "internal: pp_defn for OCaml\n"

  | Tex _ ->
      "% defns "^dc.dc_name^"\n"
      ^ (String.concat 
           "\n"
           (List.map (pp_defn m xd lookup dc.dc_wrapper) dc.dc_defns))
      ^ "\n"
      ^ "\\newcommand{"^Grammar_pp.tex_defnclass_name m dc.dc_name^"}{\n"
      ^ (String.concat 
           "\n" 
           (List.map (function d -> Grammar_pp.tex_defn_name m dc.dc_wrapper d.d_name^"{}") dc.dc_defns))
      ^ "}\n\n"

(**********************************************)
(* pp of fundefns                             *)
(**********************************************)

let pp_funclause (m:pp_mode) (xd:syntaxdefn) (fc:funclause) : string =
  let (((de1:dotenv1),de2) as de,de3,pptbe) = 
    Bounds.bound_extraction m xd fc.fc_loc [fc.fc_lhs;fc.fc_rhs] in
  let ppd_lhs = Grammar_pp.pp_symterm m xd [] de fc.fc_lhs in
  let ppd_rhs = Grammar_pp.pp_symterm m xd [] de fc.fc_rhs in
  match m with
  | Ascii ao ->
      ppd_lhs ^ " === " ^ ppd_rhs ^ "\n"
  | Tex _ ->                                  
      Grammar_pp.pp_tex_FUNCLAUSE_NAME m^"{"^ppd_lhs^"}"^"{"^ppd_rhs^"}%\n"
  | Isa _ | Hol _ | Coq _ | Caml _ | Twf _ -> 
      Auxl.error "internal: never happen in pp_funclause"

(*   | Isa _ | Hol _ | Coq _ | Twf _ -> *)
(*       let non_free_ntrs = Subrules_pp.non_free_ntrs m xd in *)

(*       (\* find all the nonterms used at non-free types *\) *)
(*       let subntrs_of_ntmvsns : (nt_or_mv*subntr_data) list -> (nontermroot * nontermroot * nonterm) list =  *)
(*         function ntmvsns -> *)
(*           Auxl.option_map  *)
(*             (function x -> match x with *)
(*             | (Ntr ntr,suffix),(None as subntr_data) ->  *)
(*                 (try   *)
(*                   let ntr_upper = List.assoc ntr non_free_ntrs in *)
(*                   Some (ntr,ntr_upper,(ntr,suffix)) *)
(*                 with *)
(*                   Not_found -> None) *)
(*             | (Ntr ntr,suffix),(Some(ntrl,ntru) as subntr_data) -> Some (ntrl,ntru,(ntr,suffix)) *)
(*             | (Mvr mvr,suffix),_ -> None)  *)
(*             ntmvsns in *)

(*       (\* first the non-list nonterms used at non-free types*\) *)
(*       let nonlist_subntrs = subntrs_of_ntmvsns de3 in *)

(*       (\* and second the list nonterms used at non-free types *\) *)
(*       let list_subntrs :((nontermroot * nontermroot * nonterm) list * string * string * string) list =  *)
(*         List.map   *)
(*           (function (bound,de1i (\*(ntmvsns,pp_squished_vars,pp_pattern,coq_type_pattern,hol_type_var)*\)) -> *)
(*             subntrs_of_ntmvsns de1i.de1_ntmvsns,de1i.de1_compound_id,de1i.de1_pattern,de1i.de1_coq_type_of_pattern) *)
(*           de1 in *)

(*       let ppd_subntrs =  *)
(*         List.map (pp_subntr m xd) nonlist_subntrs  *)
(*         @ pp_listsubntr m xd list_subntrs in *)

(*       (\* collect all the isa/coq/hol variables that should be quantified *\) *)
(*       (\* for this clause *\) *)

(*       let quantified_proof_assistant_vars =  *)
(* 	Grammar_pp.extract_quantified_proof_assistant_vars m xd de1 de2 de3 in *)

(*       (\* calculate list length constraints *\) *)
(*       (\*   There are several different things we could do here: *\) *)
(*       (\*   -a- For a proof assistant list arising from ott list forms with bounds that      *\) *)
(*       (\*        include an upper bound and (perhaps implicitly) a lower bound               *\) *)
(*       (\*        (i.e. Bound_dotform, Bound_comp_u, or Bound_comp_lu, but not Bound_comp),   *\) *)
(*       (\*        say 1..n, we could add a premise that the length of the proof assistant list*\) *)
(*       (\*        is n.                                                                       *\) *)
(*       (\*    -a.2-  It might be a nice optimisation to only include that premise if the n    *\) *)
(*       (\*        is also mentioned elsewhere in the drule.                                   *\) *)
(*       (\*   -b- For a proof assistant list arising from ott list forms with bounds that      *\) *)
(*       (\*        have a number of dots (i.e. Bound_dotform or Bound_comp_lu), say 1....n,    *\) *)
(*       (\*        we could add a premise that the length of the proof assistant list is >= 2. *\) *)
(*       (\*   -c- We could record in each Ste_list any dot length constraint from the ott      *\) *)
(*       (\*        source grammar, then add up the length constraints from the                 *\) *)
(*       (\*        symterm_list_items comprising the Ste_list, see if a non-trivial additional *\) *)
(*       (\*        constraint is required to make the Ste_list of length at least that in the  *\) *)
(*       (\*        grammar, and, if so, add that as a premise.                                 *\) *)
(*       (\*   -d- For each projection index, eg ti, we could add a constraint that i is < the  *\) *)
(*       (\*        length of the proof assistant list.                                         *\) *)
(*       (\*  *\) *)
(*       (\* Below is code to do -a-, fine for isa & hol, but not enough for Coq as there       *\) *)
(*       (\*  we need to add quantifiers for the vars that occur only in bounds                 *\) *)
(* (\*         let list_vars_with_bounds = List.map (function (b,(_,compound_ident,_,_,_)) -> compound_ident,b) de1 in *\) *)
(* (\*         let si_subtract si_lower si_upper = match si_lower,si_upper with *\) *)
(* (\*         | Si_num s, Si_var (suffv,n) ->  *\) *)
(* (\*             let n' = n + 1 - int_of_string s in *\) *)
(* (\*             if n'=0 then suffv else (match m with *\) *)
(* (\*             | Isa _ -> "("^suffv^"+("^string_of_int n'^"::nat))" *\) *)
(* (\*             | Hol _ -> "("^suffv^"+"^string_of_int n'^")" *\) *)
(* (\*             | Coq _ -> "("^suffv^"+"^string_of_int n'^")") *\) *)
(* (\*         | _,_ -> failwith ("internal error: surprising suffixes in si_subtract: "^Grammar_pp.pp_plain_suffix_item si_lower^" and "^Grammar_pp.pp_plain_suffix_item si_lower) in *\) *)
(* (\*         let lengths b = match b with *\) *)
(* (\*         | Bound_dotform bd -> Some (si_subtract bd.bd_lower bd.bd_upper), Some bd.bd_length *\) *)
(* (\*         | Bound_comp bc -> None,None *\) *)
(* (\*         | Bound_comp_u bcu -> Some (si_subtract (Si_num "0") bcu.bcu_upper), None *\) *)
(* (\*         | Bound_comp_lu bclu -> Some (si_subtract bclu.bclu_lower bclu.bclu_upper), Some bclu.bclu_length in *\) *)
(* (\*         let list_length_constraints = Auxl.option_map (function  (b,(_,compound_ident,_,coq_type_pattern,_)) ->  *\) *)
(* (\*           match lengths b with *\) *)
(* (\*           | Some s, _ ->  *\) *)
(* (\*               Some ((match m with *\) *)
(* (\*               | Isa _ -> "length" *\) *)
(* (\*               | Hol _ -> "LENGTH" *\) *)
(* (\*               | Coq _ ->  *\) *)
(* (\*                   let ty_list = Str.split (Str.regexp "(\\|*\\|)") coq_type_pattern in *\) *)
(* (\*                   "length_list_"^(String.concat "_" ty_list) *\) *)
(* (\*                   (\\* "[[TODO: insert appropriate Coq list length function for the compound_ident, built from the coq_type_pattern, here ]]" *\\)) *\) *)
(* (\*               ^ " " ^ compound_ident^" = "^s) *\) *)
(* (\*           | None,_ -> None) de1 in *\) *)

(* (\*         let ppd_premises = ppd_premises @ list_length_constraints in *\) *)
      
(*       ( match m with *)
(*       | Isa _ -> *)
(*           dr.drule_name^"I" *)
(*           ^ ": "  (\* todo - should we add the I here or before? *\) *)
(*           ^ "\""  *)
(*           ^  *)
(*             (if ppd_premises=[] && ppd_subntrs=[]  then ""  *)
(*             else  *)
(* 	      "[|" *)
(* 	      ^ (String.concat " ;\n" (ppd_subntrs @ ppd_premises))  *)
(* 	      ^ "|] ==> \n")  *)
(*           ^ ppd_conclusion  *)
(*           ^ "\"" ^ "\n" *)
(*       | Hol _ -> *)
(*           let clause_name_pp = "(clause_name \"" ^ dr.drule_name^ "\")" in *)
(*           "("  *)
(*           ^ " (\* "^dr.drule_name^" *\)  " *)
(*           ^  *)
(*             (match quantified_proof_assistant_vars with *)
(*             | [] -> "" *)
(*             | _ -> "!"  *)
(*                 ^ (let quantified_proof_assistant_vars' =  *)
(* 		  List.map  *)
(* 		    (function (var,ty,_) -> "("^var^":"^ty^")")  *)
(* 		    quantified_proof_assistant_vars in *)
(*                 String.concat " " quantified_proof_assistant_vars' ^ " . ")) *)
(*           ^  *)
(*             (if ppd_premises=[] && ppd_subntrs=[] then *)
(*               clause_name_pp^" ==> \n"  *)
(*             else  *)
(*               clause_name_pp^" /\\\n" *)
(*               ^ "(" *)
(* 	      ^ (String.concat " /\\\n"  *)
(*                    (List.map (function s->"("^s^")") (ppd_subntrs @ ppd_premises)) ) *)
(* 	      ^ ")\n ==> \n")  *)
(*           ^ "("^ppd_conclusion ^")" *)
(*           ^ ")\n\n" *)

(*       | Coq co when not co.coq_expand_lists ->  *)
(* 	  let nlh, nlhv = !(co.coq_non_local_hyp_defn), !(co.coq_non_local_hyp_defn_vars) in *)
(* 	  co.coq_non_local_hyp_defn := ""; *)
(* 	  co.coq_non_local_hyp_defn_vars := []; *)

(*           let qpavt =  *)
(* 	    List.map  *)
(* 	      (fun (v,t) ->  *)
(* 		match t with *)
(* 		| Some t -> "("^v^":"^t^")" *)
(* 		| None -> v) *)
(* 	      (Auxl.remove_duplicates ((List.map (fun (x,_,(t,_)) -> (x,Some t)) quantified_proof_assistant_vars) @ nlhv))  *)
(* 	  in *)

(*  	  " | " ^ dr.drule_name ^ " : "  *)
(* 	  ^ ( if qpavt = [] then "\n" *)
(*               else ( "forall " *)
(*                      ^ String.concat " " qpavt ^ ",\n") ) *)
(* 	  ^ ( if ppd_premises = [] && ppd_subntrs = [] && (String.compare nlh "" = 0)  *)
(*           then "" *)
(*           else begin *)
(*             nlh ^  *)
(* 	    ( if (List.length (ppd_subntrs @ ppd_premises) = 0) *)
(* 	    then "" *)
(* 	    else "     " ^ (String.concat " ->\n     " (ppd_subntrs @ ppd_premises) ^ " ->\n") )  *)
(* 	  end ) *)
(* 	  ^ "     " ^ ppd_conclusion ^ "\n" *)

(*       | Coq co when co.coq_expand_lists ->  *)

(*           let required_list_types =  *)
(*             List.map (function (_,de1i) ->  *)
(*               List.map (function ((ntmvr,_),_) ->  *)
(*                 Auxl.promote_ntmvr xd *)
(*                   (Auxl.primary_nt_or_mv_of_nt_or_mv xd ntmvr)) de1i.de1_ntmvsns) de1 in *)
          
(*           let missing_list_types =  *)
(*             List.filter  *)
(*               (fun ss ->  *)
(*                 let name = String.concat "_" (List.map (fun s -> Grammar_pp.pp_nt_or_mv_root m xd (Auxl.promote_ntmvr xd s)) ss) in *)
(*                 not (List.mem ("list_"^name) !(co.coq_list_types)))  *)
(*               required_list_types in *)

(* (\*             print_endline ("rule: " ^dr.drule_name); *\) *)
(* (\*             print_endline ("missing: "^ (String.concat " -- " *\) *)
(* (\*                                            (List.map *\) *)
(* (\*                                               (Grammar_pp.pp_nt_or_mv_root m xd) *\) *)
(* (\*                                               (List.flatten missing_list_types)))); *\) *)

(*           let (new_list_names, new_list_rules) =  *)
(*             List.split (List.map Transform.create_list_rule missing_list_types) in *)
          
(*           let rec add_list_rules_defns nln nlr = *)
(*             ( match (nln,nlr) with  *)
(*             | [],[] -> () *)
(*             | h1::t1,h2::t2 ->  *)
(*                 if not (List.mem h1 !(co.coq_list_types))  *)
(*                 then begin *)
(*                   co.coq_list_types := h1::!(co.coq_list_types); *)
(*                   co.coq_list_rules_defns := h2::!(co.coq_list_rules_defns); *)
(*                 end; *)
(*                 add_list_rules_defns t1 t2 ) in *)
          
(*           add_list_rules_defns new_list_names new_list_rules; *)

(* 	  let nlh, nlhv = !(co.coq_non_local_hyp_defn), !(co.coq_non_local_hyp_defn_vars) in *)
(* 	  co.coq_non_local_hyp_defn := ""; *)
(* 	  co.coq_non_local_hyp_defn_vars := []; *)
(* 	  let vtl = (List.map (fun (x,_,(t,_)) -> (x,Some t)) quantified_proof_assistant_vars) @ nlhv in *)
(* 	  let vtlr =  *)
(* 	    let rec check_el el l = *)
(* 	      List.mem el (List.map fst l) in *)
(* 	    let rec remove_dupl l = *)
(* 	      ( match l with *)
(* 	      | [] -> [] *)
(* 	      | (v,t)::tl ->  *)
(* 		  if check_el v tl  *)
(* 		  then remove_dupl tl *)
(* 		  else (v,t)::(remove_dupl tl) ) in *)
(* 	    remove_dupl vtl in *)
(*           let qpavt =  *)
(* 	    List.map  *)
(* 	      (fun (v,t) ->  *)
(* 		match t with *)
(* 		| Some t -> "("^v^":"^t^")" *)
(* 		| None -> v) *)
(* 	      vtlr in *)
(* (\*	    (Auxl.remove_duplicates ((List.map (fun (x,_,(t,_)) -> (x,Some t)) quantified_proof_assistant_vars) @ nlhv)) in *\) *)
(*  	  " | " ^ dr.drule_name ^ " : " ^  *)
(*           ( if qpavt = [] then "\n" *)
(*           else ( "forall " *)
(*                  ^ String.concat " " qpavt ^ ",\n") ) *)

(* 	  ^ ( if ppd_premises = [] && ppd_subntrs = [] && (String.compare nlh "" = 0)  *)
(*           then "" *)
(*           else begin *)
(*             nlh ^  *)
(* 	    ( if (List.length (ppd_subntrs @ ppd_premises) = 0) *)
(* 	    then "" *)
(* 	    else "     " ^ (String.concat " ->\n     " (ppd_subntrs @ ppd_premises) ^ " ->\n") )  *)
(* 	  end ) *)
(* 	  ^ "     " ^ ppd_conclusion ^ "\n" *)

(*       | Twf wo -> *)
(*           !(wo.twf_current_defn) ^"/"^ dr.drule_name *)
(*           ^ ": " *)
(* 	  ^ ppd_conclusion  *)
(*           ^ ""  *)
(*           ^ (if ppd_premises=[] && ppd_subntrs=[] then ""  *)
(*              else " <- " ^ (String.concat " <- " (ppd_subntrs @ ppd_premises))) *)
(*           ^ "." *)

(*       | Tex _ | Ascii _ -> raise Auxl.ThisCannotHappen  *)
(*       | Caml oo -> raise Auxl.ThisCannotHappen *)
(*        ) *)
(*   | Caml oo -> *)
(*       Auxl.error "internal: defns not implemented for OCaml" *)

let rec insert_commas l =
  match l with
  | [] -> []
  | h::[] -> h::[]
  | h::t -> h::(Hom_string ",")::(insert_commas t)

let pp_symterm_node_with_hom m xd sie de st hom = 
  match st with
  | St_node (_,stnb) -> 
      let promoted_pn = 
	try
	  let ntrt,pps = List.assoc stnb.st_rule_ntr_name xd.xd_srd.srd_subrule_pn_promotion in
	  try List.assoc stnb.st_prod_name pps with Not_found -> Auxl.error ("internal error: ppd_lhs")
	with Not_found -> stnb.st_prod_name in 
      let p = Auxl.prod_of_prodname xd promoted_pn in
      let prod_es = p.prod_es in
      let pes = Grammar_pp.pp_symterm_elements m xd [] de false prod_es stnb.st_es in
      let hom =
	match hom with
	| Some h -> h 
	| None -> 
	    let i = ref (-1) in 
	    let tmp_hom = List.map (fun _ -> i:=!i+1; Hom_index !i) pes in
	    match m with
	    | Coq _ -> Hom_string "(" :: (insert_commas tmp_hom) @ [Hom_string ")" ]
	    | Hol _ | Isa _ -> tmp_hom
	    | Caml _ | Twf _ | Ascii _ | Tex _ -> raise Auxl.ThisCannotHappen
      in String.concat " " (Grammar_pp.apply_hom_spec m xd hom pes)

  | _ -> Auxl.error "internal: pp_symterm_node_with_hom, defns.ml"

let fundefn_to_int_func (m:pp_mode) (xd:syntaxdefn) (deps:string list) (fd:fundefn) : int_func =
  let (lhs_hom, coq_indexes) = 
    match m with
    | Coq _ -> 
	(* correct if at least one argument *)
	let acc = ref [] in
	let matrix = List.map (fun c -> match c.fc_lhs with St_node (_,stnb) -> stnb.st_es) fd.fd_clauses in
	if List.length matrix = 1 
	then 
	  let l = let i = ref (-1) in List.map (fun _ -> i:=!i+1; !i) (List.hd matrix) in
	  (None,List.rev l)
	else (
	  let transpose_m = Auxl.transpose matrix in
	  let result = List.map (fun (h::l) -> List.for_all (fun e -> compare h e = 0) l) transpose_m in
	  let i = ref (-1) in
	  let tmp_hom = 
	    Auxl.option_map (fun r -> i := !i+1; if r then None else (acc := !i::!acc; Some (Hom_index !i))) result in
	  (Some ( Hom_string "(" :: (insert_commas tmp_hom) @ [Hom_string ")" ] ), !acc))
    | Hol _ | Isa _ -> (None,[]) 
    | Caml _ | Tex _ | Ascii _ | Twf _ -> raise Auxl.ThisCannotHappen in

  let header =
    match m with
    | Hol _ -> ("","","")
    | Coq _ -> 
	let prod_name = fd.fd_pn_wrapper^ fd.fd_name in
	let result_type_name = Auxl.pp_type_name m fd.fd_result_type in
        let prod_es = (Auxl.prod_of_prodname xd prod_name).prod_es in

	let struct_on = 
	  match Auxl.hom_spec_for_hom_name "coq-struct" fd.fd_homs with
	  | Some ([Hom_index i]) -> "{struct x" ^ string_of_int (i+1) ^ "} "
	  | Some _ -> 
	      Auxl.warning "malformed coq-struct homomorphism"; 
	      "{struct <<<malformed term in coq-struct hom>>>}"
	  | None -> "" in

	let (type_defn,match_string) =
          let ss = (Auxl.option_map (Grammar_pp.pp_element m xd []) prod_es) in

          match ss with
          | [] -> (" :=","")      (* FZ THIS IS PROBABLY WRONG *)
          | [s] -> ("(x1:"^s^") ", "  match x1 with\n")
          | _ -> 
	      let i = ref 0 in
	      let v = 
		String.concat " "
		  (List.map (fun s -> "(x"^(i := !i+1; string_of_int !i)^":"^s^")") ss) in
	      let ms = 
		"  match ("
		^ String.concat "," (List.map (fun i -> "x"^string_of_int (i+1)) (List.rev coq_indexes))
		^ ") with\n" in
	      (v,ms) 
	in
	( fd.fd_name ^ " " ^ type_defn ^ " "^struct_on^": " ^ result_type_name ^ ":=\n",
	  "",
	  match_string)
    | Isa _ ->
	let isa_type_of_fundefn : syntaxdefn -> fundefn -> string =
          fun xd fd ->
            (* seems simplest to find the type associated with the
               production that we added to the language for this defn,
               rather than build a type directly from the defn. Though
               (now wrappers are working properly) that does mean we
               need to pass them in. *) 

	    let result_type_name = Auxl.pp_type_name m fd.fd_result_type in
            let prod_name = fd.fd_pn_wrapper ^ fd.fd_name in
            let es = (Auxl.prod_of_prodname xd prod_name).prod_es in
            let ss = (Auxl.option_map (Grammar_pp.pp_element m xd []) es)in
            (match ss with
            | [] -> "unit"
            | [s] -> s
            | _ -> "" ^ String.concat " => " ss ^"") ^ " => " ^ result_type_name in

	let h1 = fd.fd_name ^ " :: \"" ^ isa_type_of_fundefn xd fd ^"\"\n" in

	(h1,"","")

(*         (let fancy_syntax_clauses = *)
(*           Auxl.option_map *)
(* 	    (Grammar_pp.isa_fancy_syntax_clause_for_fundefn m xd io fdc.fdc_name) *)
(* 	    fdc.fdc_fundefns in *)
(*         match fancy_syntax_clauses with *)
(*         | [] -> "" *)
(*         | fscs -> *)
(*             "syntax\n" *)
(*             ^ String.concat "\n" (List.map fst fscs) *)
(*             ^ "\n\ntranslations\n" *)
(*             ^ String.concat "\n" (List.map snd fscs) *)
(*             ^ "\n\n") *)
(*       ^ "inductive " *)
(*       ^ (String.concat *)
(*            " " *)
(*            (List.map (function fd ->  fd.fd_name) fdc.fdc_fundefns)) *)
(*       ^ "\nintros\n\n" *)
    | Caml _ | Tex _ | Ascii _ | Twf _ -> raise Auxl.ThisCannotHappen 

  in
  
  
  let clauses =
    List.map 
      (fun fc ->
	let (((de1:dotenv1),de2) as de,de3,pptbe) = 
	  Bounds.bound_extraction m xd fc.fc_loc [fc.fc_lhs;fc.fc_rhs] in
	let ppd_lhs =
	  pp_symterm_node_with_hom m xd [] de fc.fc_lhs lhs_hom in

	let ppd_rhs = Grammar_pp.pp_symterm m xd [] de fc.fc_rhs in
	("", ppd_lhs, ppd_rhs))
      fd.fd_clauses in

  let deps =
    match m with
    | Isa _ | Hol _ -> deps
    | Coq _ -> 
	( match Auxl.hom_spec_for_hom_name "coq-struct" fd.fd_homs with
	| Some _ -> deps
	| None -> List.filter (fun s -> not (String.compare s fd.fd_name = 0)) deps )
    | Caml _ | Tex _ | Ascii _ | Twf _ -> raise Auxl.ThisCannotHappen 
  in

  { r_fun_id = fd.fd_name;
    r_fun_dep = deps;
    r_fun_type = "TODO";  
    r_fun_header = header;
    r_fun_clauses = clauses
  }

let pp_fundefn (m:pp_mode) (xd:syntaxdefn) lookup (fd:fundefn) : string =
  match m with
  | Ascii _ ->
      Grammar_pp.pp_FUNDEFN ^"\n"
      ^ Grammar_pp.pp_symterm m xd [] de_empty fd.fd_form
      ^ Grammar_pp.pp_COLONCOLON ^" "^fd.fd_result^" "
      ^ Grammar_pp.pp_COLONCOLON ^" "^fd.fd_name^" "
(*      ^ Grammar_pp.pp_COLONCOLON ^" "^fd.fd_pn_wrapper^" " *)
      ^ Grammar_pp.pp_BY
      ^ "\n\n"
      ^ String.concat "\n" (List.map (pp_funclause m xd) fd.fd_clauses)

  | Tex xo ->
      let prod_name = (fd.fd_pn_wrapper^fd.fd_name) in
      let p = Auxl.prod_of_prodname xd prod_name in
      let es = p.prod_es in 
      let pp_com = Grammar_pp.pp_com_es m xd fd.fd_homs es  in
      "% fundefn "^fd.fd_name^"\n"
      ^ "\n\\newcommand{"^Grammar_pp.tex_fundefn_name m fd.fd_name^"}[1]{"
      ^ "\\begin{"^Grammar_pp.pp_tex_FUNDEFN_BLOCK_NAME m ^"}[#1]{$" 
      ^ Grammar_pp.pp_symterm m xd [] de_empty fd.fd_form 
      ^ "$}{"
      ^ pp_com
      ^ "}\n"
      ^ String.concat "" 
          (List.map (pp_funclause m xd) fd.fd_clauses)
      ^ "\\end{"^Grammar_pp.pp_tex_FUNDEFN_BLOCK_NAME m ^"}" 
      ^ "}\n\n"

  | Isa _ | Hol _ | Coq _ | Twf _ -> 
      Auxl.error "internal: never happen in pp_funclause"
  | Caml oo ->
      Auxl.error "internal: funcs not implemented for OCaml"

let pp_fundefnclass (m:pp_mode) (xd:syntaxdefn) lookup (fdc:fundefnclass) : string =
  match m with
  | Ascii _ -> 
      Grammar_pp.pp_FUNDEFNCLASS ^"\n"
      ^ fdc.fdc_name ^" "
      ^ Grammar_pp.pp_CCE^"\n\n"
      ^ String.concat "\n\n" (List.map (pp_fundefn m xd lookup) fdc.fdc_fundefns)

  | Tex _ ->
      "% fundefns "^fdc.fdc_name^"\n"
      ^ (String.concat 
           "\n"
           (List.map (pp_fundefn m xd lookup) fdc.fdc_fundefns))
      ^ "\n"
      ^ "\\newcommand{"^Grammar_pp.tex_fundefnclass_name m fdc.fdc_name^"}{\n"
      ^ (String.concat 
           "\n" 
           (List.map (function fd -> Grammar_pp.tex_fundefn_name m fd.fd_name^"{}") fdc.fdc_fundefns))
      ^ "}\n\n"
 
  | Isa _ | Coq _ | Hol _ ->
      let proof = 
	let pp_proof h =
	  ( match h with 
	  | Some ([Hom_string s]) -> Some s
	  | None -> None
	  | _ -> Auxl.warning "malformed isa-proof/hol-proof hom"; Some "<<<malformed isa-proof/hol-proof hom>>>" ) in 
	( match m with
	| Coq _ -> None
	| Isa _ -> pp_proof (Auxl.hom_spec_for_hom_name "isa-proof" fdc.fdc_homs) 
	| Hol _ -> pp_proof (Auxl.hom_spec_for_hom_name "hol-proof" fdc.fdc_homs)
	| _ -> raise Auxl.ThisCannotHappen ) in
      let deps = List.map (fun fd -> fd.fd_name) fdc.fdc_fundefns in
      let int_funcs = 
	{ i_funcs = List.map (fundefn_to_int_func m xd deps) fdc.fdc_fundefns;
	  i_funcs_proof = proof } in
      let int_funcs_collapsed = Dependency.collapse m xd int_funcs in
      Auxl.print_with_comment m "\n" ("funs "^fdc.fdc_name) 
	(Dependency.compute m xd int_funcs_collapsed)
	
  | Caml _ | Twf _ -> Auxl.warning "internal: fundefnclass not implemented for Ocaml / Twelf"; ""

(*   | Isa io -> *)
(*       let isa_type_of_fundefn : syntaxdefn -> fundefn -> string = *)
(*         fun xd fd -> *)
(*         (\* seems simplest to find the type associated with the production that *)
(*            we added to the language for this defn, rather than build a type *)
(*            directly from the defn. Though (now wrappers are working properly) *)
(*            that does mean we need to pass them in. *\) *)
(*           let result_type_name = Auxl.pp_type_name m fd.fd_result_type in *)

(*           let prod_name = fd.fd_pn_wrapper ^ fd.fd_name in *)
(*           let es = (Auxl.prod_of_prodname xd prod_name).prod_es in *)
(*           let ss = (Auxl.option_map (Grammar_pp.pp_element m xd []) es)in *)

(*           (match ss with *)
(*           | [] -> "unit" *)
(*           | [s] -> s *)
(*           | _ -> ""^String.concat "=>" ss ^"") ^ " => " ^ result_type_name in *)
(* (\*             match Grammar_pp.pp_elements m xd es true with  *\) *)
(* (\*               None -> "unit"  (\* TODO is that sensible? *\) *\) *)
(* (\*             | Some s -> s in *\) *)
(*       "(\*" ^ "funs "^fdc.fdc_name^" *\)\n" *)
(*       ^ "consts\n  " *)
(*       ^ (String.concat *)
(*            "\n  " *)
(*            (List.map *)
(*               (function fd -> *)
(*                 fd.fd_name *)
(*                 ^ " :: \"" *)
(*                   ^ isa_type_of_fundefn xd fd *)
(*                 ^"\"") *)
(*               fdc.fdc_fundefns)) *)
(*       ^ "\n\n" *)
(*       ^ *)
(*         (let fancy_syntax_clauses = *)
(*           Auxl.option_map *)
(* 	    (Grammar_pp.isa_fancy_syntax_clause_for_fundefn m xd io fdc.fdc_name) *)
(* 	    fdc.fdc_fundefns in *)
(*         match fancy_syntax_clauses with *)
(*         | [] -> "" *)
(*         | fscs -> *)
(*             "syntax\n" *)
(*             ^ String.concat "\n" (List.map fst fscs) *)
(*             ^ "\n\ntranslations\n" *)
(*             ^ String.concat "\n" (List.map snd fscs) *)
(*             ^ "\n\n") *)
(*       ^ "inductive " *)
(*       ^ (String.concat *)
(*            " " *)
(*            (List.map (function fd ->  fd.fd_name) fdc.fdc_fundefns)) *)
(*       ^ "\nintros\n\n" *)
(*       ^ (String.concat "\n" (List.map (pp_fundefn m xd lookup) fdc.fdc_fundefns)) *)

(*   | Hol ho -> *)
(*       "(\* funs "^fdc.fdc_name^" *\)\n\n" *)
(*       ^ "val ("^ fdc.fdc_name^"_rules" *)
(*       ^ ", "   ^ fdc.fdc_name^"_ind" *)
(*       ^ ", "   ^ fdc.fdc_name^"_cases" *)
(*       ^ ") = Hol_reln `\n" *)
(*       ^ (String.concat "/\\\n" *)
(* 	   (List.map (pp_fundefn m xd lookup) fdc.fdc_fundefns)) *)
(*       ^ "`;\n" *)



 (* ************************************************ *)

let pp_fun_or_reln_defnclass m xd lookup frdc = match frdc with
| FDC fdc -> pp_fundefnclass m xd lookup fdc
| RDC dc -> pp_defnclass m xd lookup dc

let pp_fun_or_reln_defnclass_list : pp_mode -> syntaxdefn -> made_parser -> fun_or_reln_defnclass list -> string
    = fun m xd lookup frdcs -> 
      if Auxl.uses_lists_syntaxdefn xd > 0 then begin
	Substs_pp.pp_list_mono_lemmas m;
	Substs_pp.pp_list_all_cong_lemma m;
	Substs_pp.pp_list_simp_lemmas m xd end;
      match m with
      | Ascii _ | Isa _ | Hol _ -> 
	  if frdcs = [] then "" else
	  ( "(** definitions *)\n"
            ^ String.concat "\n" (List.map (pp_fun_or_reln_defnclass m xd lookup) frdcs) )
      | Twf _ -> 
	  if frdcs = [] then "" else
	  ( "%%% definitions %%%\n\n"
            ^ String.concat "\n" (List.map (pp_fun_or_reln_defnclass m xd lookup) frdcs) )
      | Coq co ->
          let defns_list = 
	    if frdcs = [] then "" 
            else
	      ( "(** definitions *)\n"
                ^ String.concat "\n" (List.map (pp_fun_or_reln_defnclass m xd lookup) frdcs) ) in
          let aux_rules_defns =
            if !(co.coq_list_rules_defns) = []
            then ""
            else
	      let s = String.concat "\n" !(co.coq_list_rules_defns) in
	      co.coq_list_rules_defns := [];
              "\n(* auxiliary list types for defns *)\n"
              ^ s
              ^ "\n" in
	  aux_rules_defns ^ defns_list
      | Tex _ -> 
          "% defnss\n"
          ^ String.concat "\n" (List.map (pp_fun_or_reln_defnclass m xd lookup) frdcs)
          ^ "\\newcommand{"^Grammar_pp.pp_tex_DEFNSS_NAME m^"}{\n"
          ^ String.concat "\n" (List.map (function frdc -> match frdc with FDC fdc -> Grammar_pp.tex_fundefnclass_name m fdc.fdc_name | RDC dc -> Grammar_pp.tex_defnclass_name m dc.dc_name) frdcs)
          ^ "}\n\n" 
      | Caml _ ->
	  "(* defns not defined in OCaml output *)"


(** *********************** *)
(** processing of raw defns *)
(** *********************** *)


let fakesucc = let n=ref 0 in function () -> n:= !n+1; 
  Auxl.string_of_char_list
    (List.map 
       (function c -> char_of_int (int_of_char c + 49)) 
       (Auxl.char_list_of_string 
          (let s0=string_of_int !n in
          (String.make (6-String.length s0) '0' ^ s0))))


let process_semiraw_rule : pp_mode -> syntaxdefn -> made_parser -> string -> string -> string -> semiraw_rule -> processed_semiraw_rule option 
    = fun m xd lookup defnclass_wrapper defn_wrapper prod_name -> function sr -> match sr with

      Lined_rule(l0,lss1,ls,lss2) ->

        let (l,s)=ls in 
        let lexbuf = Lexing.from_string s in
        let annot : raw_drule_line_annot = try
          Grammar_parser.drule_line_annot (Grammar_lexer.my_lexer true Grammar_lexer.metalang)  lexbuf 
        with 
        |  Parsing.Parse_error | My_parse_error _ ->
            raise (Rule_parse_error ("bad annotation in \""^s^"\" at "^Location.pp_loc l))
        |  e ->
            (print_string ("exception in parsing \""^s^"\" at "^Location.pp_loc l^"\n");
             flush stdout;
             raise e) in
        (* let categories = List.map del c in *)

        let premises = List.map (function (l,s)->Term_parser.just_one_parse m xd lookup rn_formula false l s) lss1 in
        let conclusion =
          match lss2 with
          | [] -> raise (Rule_parse_error ("rule with no conclusion at " ^ Location.pp_loc l))
          | [(l,s)] ->
              (Term_parser.just_one_parse m xd lookup rn_formula false l) 
                (Auxl.psuedo_terminal_of_prod_name prod_name 
                 ^ " " ^ s)
          | _ -> raise (Rule_parse_error ("rule with multiple conclusions at " ^ Location.pp_loc l))
        in
        let c = Term_parser.cd_env_of_syntaxdefn xd in
        let dr = {drule_name = defnclass_wrapper^defn_wrapper^annot.dla_name;
                  drule_categories =
                  List.fold_left (fun s (_,c) -> StringSet.add c s)
                    StringSet.empty annot.dla_categories;
                  drule_premises = premises;
                  drule_conclusion = conclusion;
                  drule_homs = List.map (Grammar_typecheck.cd_hom Hu_drule c []) annot.dla_homs ;
                  drule_loc = l0} in
        (* print_string (pp_ascii_drule xd dr^"\n\n"); *)
        Some (PSR_Rule dr)

	  (* hacky test harness code for now*)
    | Lineless_rule (l0,lss) -> 
        (match lss with
        | [] -> None
        | _ -> let dr = {drule_name = "fake" ^ fakesucc ();
                         drule_categories = StringSet.singleton "F";
                         drule_premises = [];
                         drule_conclusion = (Term_parser.just_one_parse m xd lookup rn_formula false) l0 (String.concat " " (List.map snd lss));
                         drule_homs = [];  
                         drule_loc = l0} in
          Some(PSR_Rule dr))
    | Defncom (l0,es) ->
        Some(PSR_Defncom es)

let process_raw_defn : pp_mode -> syntaxdefn -> made_parser -> string -> string -> raw_defn -> defn
    = fun m xd lookup defnclass_name defnclass_wrapper rd -> 

      let prod_name = (defnclass_wrapper^rd.raw_d_name) in
      let p = Auxl.prod_of_prodname xd prod_name in
      let es = p.prod_es in 
      (match m with Ascii ao when ao.ppa_show_defns -> 
        print_string (rd.raw_d_name^" ");flush stdout 
      | _ -> ());
      let c = Term_parser.cd_env_of_syntaxdefn xd in
      { d_name=rd.raw_d_name;
        d_form= 
        (let reconstructed_string = 
          Auxl.psuedo_terminal_of_prod_name prod_name ^" "
          ^ (String.concat " " 
               (List.map Grammar_pp.pp_raw_element rd.raw_d_es)) in
        (Term_parser.just_one_parse m xd lookup defnclass_name false rd.raw_d_loc) reconstructed_string);
        (* d_categories = List.map (Location.del) rd.raw_d_categories; *)
        d_wrapper= rd.raw_d_wrapper;
        d_rules= Auxl.option_map
          (process_semiraw_rule m xd lookup defnclass_wrapper rd.raw_d_wrapper prod_name) 
          rd.raw_d_body;
        d_homs =  List.map (Grammar_typecheck.cd_hom Hu_defn c es) rd.raw_d_homs;
        d_loc = dummy_loc} 

let process_raw_defnclass : pp_mode -> syntaxdefn -> made_parser -> raw_defnclass -> defnclass
    = fun m xd lookup rdc -> 
      {dc_name=rdc.raw_dc_name;
       dc_language_names = []; (* TODO *)
       dc_wrapper = rdc.raw_dc_wrapper;
       dc_defns = List.map 
         (process_raw_defn m xd lookup rdc.raw_dc_name rdc.raw_dc_wrapper) rdc.raw_dc_defns;
       dc_loc = dummy_loc} 

(* ******************** *)

let process_raw_funclause : pp_mode -> syntaxdefn -> made_parser -> string -> string -> string -> nontermroot-> raw_funclause -> funclause option 
    = fun m xd lookup fundefnclass_name pn_wrapper fundefn_name result_type -> function (l,s) ->
(*
      print_string ("funclause prodname = "^prodname^"\n");flush stdout;
      print_string ("s = "^s^"\n");flush stdout;
      let fundefnclassname = "FUNTODO" in
*)
      let st = Term_parser.just_one_parse m xd lookup fundefnclass_name false l s in
      
      match st with
      | (St_node(_,{st_es=es })) -> 
          let e,es' = 
            (let rev_es = List.rev es in
            match rev_es with
            | e::es' -> e, List.rev es'
            | _ -> failwith "process_raw_funclause internal error - empty list ") in
          (* synthesise a symterm node for the meta production of the term grammar *)
          let lhs = St_node(dummy_loc, { st_rule_ntr_name=result_type;
                                         st_prod_name= pn_wrapper^fundefn_name;
                                         st_es = es';
                                         st_loc = dummy_loc }) in
          let rhs = 
	    match e with 
	    | Ste_st(_,st) -> st 
	    | _ -> Auxl.error "process_raw_funclause internal error - bad rhs" in
(*       print_string ("lhs symterm: "^ Grammar_pp.pp_plain_symterm lhs ^ "\n");flush stdout;  *)
(*       print_string ("rhs symterm: "^ Grammar_pp.pp_plain_symterm rhs ^ "\n");flush stdout;  *)

          Some 
            { fc_lhs = lhs; 
              fc_rhs = rhs; 
              fc_loc = l }
      | _ -> 
	  Auxl.warning ("process_raw_funclause lost symterm: "
			^ Grammar_pp.pp_plain_symterm st ^ "\n"); 
	  None
          
let process_raw_fundefn : pp_mode -> syntaxdefn -> made_parser -> string -> raw_fundefn -> fundefn
    = fun m xd lookup fundefnclass_name rfd -> 

      let prod_name = (rfd.raw_fd_pn_wrapper^rfd.raw_fd_name) in
      let p = Auxl.prod_of_prodname xd prod_name in
      let es = p.prod_es in 
(*       (match m with Ascii ao when ao.ppa_show_defns ->  *)
(*         print_string (rfd.raw_fd_name^" ");flush stdout  *)
(*       | _ -> ()); *)
      (*print_string ("process_raw_fundefn prod_name = "^prod_name^"***"^(String.concat "" (List.map Grammar_pp.pp_plain_element es))^"***\n");flush stdout;  *)
      let c = Term_parser.cd_env_of_syntaxdefn xd in
      let fd_result = snd rfd.raw_fd_result in
      let fd_result_type = rfd.raw_fd_result_type in
      { fd_name=rfd.raw_fd_name;
        fd_form= 
        (let reconstructed_string = 
          Auxl.psuedo_terminal_of_prod_name prod_name ^" "
          ^ (String.concat " " 
               (List.map Grammar_pp.pp_raw_element rfd.raw_fd_es)) in
(*         print_string ("reconstructed string = "^reconstructed_string^"\n");flush stdout;*)
        (Term_parser.just_one_parse m xd lookup fd_result_type false rfd.raw_fd_loc) reconstructed_string);
        (* d_categories = List.map (Location.del) rd.raw_d_categories; *)
        fd_result = fd_result; 
        fd_result_type = fd_result_type; 
        fd_pn_wrapper = rfd.raw_fd_pn_wrapper;
        fd_clauses = Auxl.option_map
          (process_raw_funclause m xd lookup fundefnclass_name rfd.raw_fd_pn_wrapper rfd.raw_fd_name fd_result_type) 
          rfd.raw_fd_clauses;
        fd_homs =  List.map (Grammar_typecheck.cd_hom Hu_fundefn c es) rfd.raw_fd_homs;
        fd_loc = dummy_loc} 

let process_raw_fundefnclass : pp_mode -> syntaxdefn -> made_parser -> raw_fundefnclass -> fundefnclass
    = fun m xd lookup rfdc -> 
      let c = Term_parser.cd_env_of_syntaxdefn xd in
      { fdc_name=rfdc.raw_fdc_name;
 (*       fdc_language_names = []; (* TODO *) *)
        fdc_homs =  List.map (Grammar_typecheck.cd_hom Hu_fundefnclass c []) rfdc.raw_fdc_homs;
	fdc_fundefns = List.map 
          (process_raw_fundefn m xd lookup rfdc.raw_fdc_name) rfdc.raw_fdc_fundefns;
	fdc_loc = dummy_loc} 

(* **************** *)        

let process_raw_defnclasss : pp_mode -> syntaxdefn -> made_parser -> raw_fun_or_reln_defnclass list -> fun_or_reln_defnclass list 
    = fun m xd lookup rdcs ->
      let dcs = List.map (function rfrdc -> 
        let frdc = 
          match rfrdc with
          | Raw_FDC rfdc -> FDC (process_raw_fundefnclass m xd lookup rfdc)
          | Raw_RDC rdc -> RDC (process_raw_defnclass m xd lookup rdc) in
        (match m with 
        | Ascii ao when ao.ppa_show_defns -> 
            print_string 
              (pp_fun_or_reln_defnclass m xd lookup frdc)
        | _ -> ());
        flush stdout;
        frdc
                         ) rdcs in
      
(*      print_string (let m = pp_syntaxdefn_opts_default in String.concat "\n" (List.map (function dc -> pp_ascii_defnclass m xd dc (* ^pp_isa_defnclass m xd dc*)) dcs)); *)
      dcs
	(* raise NotImplementedYet *)


(** ***************** *)
(** goodness counting *)
(** ***************** *)

(* check that a symterm has no uninterpreted elements *)
let rec no_uninterpreted_symterm st 
    = match st with
    | St_node (l,stnb) -> no_uninterpreted_symterm_node_body stnb
    | St_nonterm (l,ntrp,nt)  -> true
    | St_nontermsub (l,ntrpl,ntrpt,nt)  -> true
    | St_uninterpreted (l,s) -> false
and no_uninterpreted_symterm_element e 
    = match e with
    | Ste_st (l,st) -> no_uninterpreted_symterm st
    | Ste_metavar (l,mvrp,mb) -> true
    | Ste_var (l,mvrp,var) -> true
    | Ste_list (l,stlis) -> List.for_all no_uninterpreted_symterm_list_item stlis
and no_uninterpreted_symterm_list_item stli 
    = match stli with
    | Stli_single(l,stes) -> 
        List.for_all no_uninterpreted_symterm_element stes
    | Stli_listform stlb ->
        no_uninterpreted_symterm_list_body stlb
and no_uninterpreted_symterm_node_body stnb
    = List.for_all no_uninterpreted_symterm_element stnb.st_es
and no_uninterpreted_symterm_list_body stlb
    = List.for_all no_uninterpreted_symterm_element stlb.stl_elements
    
(* count the number of rules that do not / do have uninterpreted elements *)
let sum_pairs xys = 
  List.fold_left (function (x,y) -> function (x',y') -> (x+x',y+y')) (0,0) xys 

let count_good_clauses_drule : drule -> int*int
    = fun dr -> 
      let sts = dr.drule_conclusion::dr.drule_premises in
      let n_all = List.length sts in
      let n_good = 
	List.length ((List.filter no_uninterpreted_symterm) sts) in
      (n_good, n_all-n_good)

let count_good_clauses_psr : processed_semiraw_rule -> int*int
    = fun psr -> match psr with
    | PSR_Rule dr -> count_good_clauses_drule dr
    | PSR_Defncom _ -> (0,0)

let count_good_clauses_defn : defn -> int*int
    = fun d -> 
      sum_pairs (List.map count_good_clauses_psr d.d_rules)

let count_good_clauses_defnclass : defnclass -> int*int
    = fun dc ->
      sum_pairs (List.map count_good_clauses_defn dc.dc_defns)

let count_good_clauses_defnclasses : defnclass list-> int*int
    = fun dcs ->
      sum_pairs (List.map count_good_clauses_defnclass dcs)

let count_good_drule : drule -> int*int
    = fun dr -> 
      let (n_good,n_bad) = count_good_clauses_drule dr in 
      if n_bad=0 then (1,0) else (0,1)                            
let count_good_psr : processed_semiraw_rule -> int*int
    = fun psr ->
      match psr with
      | PSR_Rule dr -> count_good_drule dr
      | PSR_Defncom es -> (0,0)
let count_good_defn : defn -> int*int
    = fun d -> 
      sum_pairs (List.map count_good_psr d.d_rules)
let count_good_defnclass : defnclass -> int*int
    = fun dc ->
      sum_pairs (List.map count_good_defn dc.dc_defns)
let count_good_defnclasses : defnclass list-> int*int
    = fun dcs ->
      sum_pairs (List.map count_good_defnclass dcs)

let pp_counts sd = 
  let defnclasses = Auxl.option_map (fun frdc -> match frdc with RDC dc -> Some dc | FDC _ -> None) sd.relations in
  let (good,bad) = count_good_defnclasses defnclasses in
  let (good',bad') = count_good_clauses_defnclasses defnclasses in
  bad <> 0 || bad' <> 0,
  "Definition rules:        "
  ^string_of_int good ^" good    "
  ^string_of_int bad^" bad\n"
  ^ "Definition rule clauses: "
  ^string_of_int good'^" good    "
  ^string_of_int bad'^" bad\n"
                         




